package com.edb;
import java.sql.*;
import java.util.*;
import java.io.*;
import java.util.concurrent.TimeUnit;
import java.nio.ByteBuffer;
import org.postgresql.replication.ReplicationSlotInfo;
import org.postgresql.replication.LogSequenceNumber;
import org.postgresql.replication.PGReplicationStream;
import org.postgresql.PGConnection;
import org.postgresql.PGProperty;



/**
 * Create replication slot on standby server
 */
public class Nickel {

	private String m_primaryConStr = "";
	private String m_primaryUserName = "";
	private String m_primaryUserPassword = "";
	private String m_repSlotName = "";
	private String m_pubName = "";
	private Connection m_primaryConnection = null;
	private long m_standbyBackendPID = 0;
	private LogSequenceNumber m_slotLSN = null;
	private String m_slotSnapshotID = null;

	Nickel(String primaryConStr, String primaryUserName, String primaryUserPassword, String slotName, String pubName) {
		this.m_primaryConStr = primaryConStr;
		this.m_primaryUserName = primaryUserName;
		this.m_primaryUserPassword = primaryUserPassword;
		this.m_repSlotName = slotName;
		this.m_pubName = pubName;
	}

	private boolean connectWithPrimary() {
		Properties props = new Properties();
		PGProperty.USER.set(props, this.m_primaryUserName);
		PGProperty.PASSWORD.set(props, this.m_primaryUserPassword);
		PGProperty.ASSUME_MIN_SERVER_VERSION.set(props, "9.6");
		PGProperty.REPLICATION.set(props, "database");
		PGProperty.PREFER_QUERY_MODE.set(props, "simple");

		try {
			Class.forName("org.postgresql.Driver");
		}
		catch(ClassNotFoundException e) {
			System.out.println("Class Not Found : " + e.getMessage());
			return false;
		}

		try {
			m_primaryConnection = DriverManager.getConnection(this.m_primaryConStr, props);
		} catch (SQLException e) {
			System.err.format("SQL State: %s\n%s", e.getSQLState(), e.getMessage());
			return false;
		} catch (Exception e) {
			e.printStackTrace();
			return false;
		}
		return true;
	}

	public void startWALParser() {
		try {
			PGConnection replConnection = m_primaryConnection.unwrap(PGConnection.class);
			ReplicationSlotInfo info = replConnection
							.getReplicationAPI()
							.createReplicationSlot()
							.logical()
							.withSlotName(m_repSlotName)
							.withOutputPlugin("pgoutput")
							.make();

			m_slotLSN = info.getConsistentPoint();
			m_slotSnapshotID = info.getSnapshotName();
			
			PGReplicationStream stream =
						replConnection.getReplicationAPI()
						.replicationStream()
						.logical()
						.withSlotName(m_repSlotName)
						.withSlotOption("proto_version", 1)
						.withSlotOption("publication_names", "barium")
						.withStatusInterval(120, TimeUnit.SECONDS)
						.start();

			while (true) {
				//non blocking receive message
				ByteBuffer msg = stream.readPending();

				if (msg == null) {
					TimeUnit.MILLISECONDS.sleep(1000L);
					continue;
				}

				int offset = msg.arrayOffset();
				byte[] source = msg.array();
				int length = source.length - offset;
				System.out.println("\t[SCT]:: Length " + length);
				System.out.println();
				int cnt = 0;
				for (byte b : source) {
					cnt++;
					String st = String.format("%02X ", b);
					System.out.print(st);
					if (cnt > 70) {
						cnt = 0;
						System.out.println();
					}
				}
				System.out.println();

				cnt = 0;
				for (byte b : source) {
					cnt++;
					if (b >= 32 && b < 127) {
						String st = String.format(" %c ", b);
						System.out.print(st);
					} else {
						String st = String.format("   ", b);
						System.out.print(st);
					}

					if (cnt > 70) {
						cnt = 0;
						System.out.println();
					}
				}
				System.out.println();

			  //feedback
			  stream.setAppliedLSN(stream.getLastReceiveLSN());
			  stream.setFlushedLSN(stream.getLastReceiveLSN());
			}

		} catch (SQLException e) {
			System.err.format("\t[SCT]: SQL State: %s\n%s", e.getSQLState(), e.getMessage());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private long getBackendPID() {
		String cmd = "SELECT pg_backend_pid() as pid;";
		long pid = 0;
		if (m_primaryConnection == null)
			return 0;
		try {
			PreparedStatement preparedStatement = m_primaryConnection.prepareStatement(cmd);
			System.out.println("[MAIN]: Going to get backend PID ");
			ResultSet resultSet = preparedStatement.executeQuery();
			boolean bret = resultSet.next();
			if (bret) {
				pid = resultSet.getLong("pid");
			}
			preparedStatement.close();
		} catch (SQLException e) {
			System.err.format("[MAIN]: SQL State: %s\n%s", e.getSQLState(), e.getMessage());
			return 0;
		} catch (Exception e) {
			e.printStackTrace();
			return 0;
		}
		return pid;
	}
	
    public boolean startStreaming()
    {
		System.out.println("[MAIN]: Going to connect with primary .....");
		connectWithPrimary();
		long walSenderPID = getBackendPID();
		System.out.println("[MAIN]: WAL Sender PID is " + walSenderPID);
		
		startWALParser();

		if (m_primaryConnection != null)
		{
			try {
				m_primaryConnection.close();
			} catch (SQLException e) {
			} catch (Exception e) {
			}
		}
        System.out.println("[MAIN]: All Done");
        return true;
    }
}

